class InvalidEndpointTypeException(Exception):
    def __init__(self):
        super().__init__("Invalid endpoint type. Must be 'handler' or 'background'")


class RouteAlreadyInUseException(Exception):
    def __init__(self):
        super().__init__("Route already in use")





from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, Optional, Union, Generator, Optional, Union
import json as jsonlib

class RequestHeaders():
    def __init__(self, headers: Dict[str, str]):
        self._headers = {}
        for key in headers:
            self._headers[self._normalize_key(key)] = headers[key]

    def _normalize_key(self, key):
        if not isinstance(key, str):
            raise KeyError(key)
        return key.upper().replace("-", "_")

    def __getitem__(self, key):
        print(self._headers)
        key = self._normalize_key(key)
        return self._headers[key]

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

@dataclass
class Request():
    id: str
    headers: RequestHeaders
    json: Dict[str, Any]

ResponseBody = Union[bytes, Generator[bytes, None, None]]
RequestID = str

class Response():
    def __init__(self, status: int = 200, json: Optional[dict] = None, headers: Optional[dict] = None, body: Optional[ResponseBody] = None):
        assert json == None or body == None, "Potassium Response object cannot have both json and body set"


        self.headers = headers if headers != None else {}

        # convert json to body if not None
        if json != None:
            self.body = jsonlib.dumps(json).encode("utf-8")
            self.headers["Content-Type"] = "application/json"
        else:
            self.body = body

        self.status = status

    @property
    def json(self):
        if self.body == None:
            return None
        if type(self.body) == bytes:
            try:
                return jsonlib.loads(self.body.decode("utf-8"))
            except:
                return None
        return None
            
    @json.setter
    def json(self, json):
        self.body = jsonlib.dumps(json).encode("utf-8")
        self.headers["Content-Type"] = "application/json"





from enum import Enum
import time
from typing import List, Tuple
from dataclasses import dataclass

from .types import RequestID

class InvalidStatusEvent(Exception):
    pass

class StatusEvent(Enum):
    INFERENCE_REQUEST_RECEIVED = "INFERENCE_REQUEST_RECEIVED"
    INFERENCE_START = "INFERENCE_START"
    INFERENCE_END = "INFERENCE_END"
    WORKER_STARTED = "WORKER_STARTED"
    BAD_REQUEST_RECEIVED = "BAD_REQUEST_RECEIVED"

@dataclass
class PotassiumStatus():
    """PotassiumStatus is a simple class that represents the status of a Potassium app."""
    num_started_inference_requests: int
    num_completed_inference_requests: int
    num_bad_requests: int
    num_workers: int
    num_workers_started: int
    idle_start_timestamp: float
    in_flight_request_start_times: List[Tuple[RequestID, float]]

    @staticmethod
    def initial(num_workers: int) -> "PotassiumStatus":
        return PotassiumStatus(
            num_started_inference_requests=0,
            num_completed_inference_requests=0,
            num_bad_requests=0,
            num_workers=num_workers,
            num_workers_started=0,
            idle_start_timestamp=time.time(),
            in_flight_request_start_times=[]
        )

    @property
    def requests_in_progress(self):
        return self.num_started_inference_requests - self.num_completed_inference_requests

    @property
    def gpu_available(self):
        if self.num_workers_started < self.num_workers:
            return False
        return self.num_workers - self.requests_in_progress > 0

    @property
    def sequence_number(self):
        return self.num_started_inference_requests + self.num_bad_requests

    @property
    def idle_time(self):
        num_received_requests_not_completed = self.num_started_inference_requests - self.num_completed_inference_requests
        has_incomplete_requests = num_received_requests_not_completed > 0
        if not self.gpu_available or has_incomplete_requests:
            return 0
        return time.time() - self.idle_start_timestamp

    @property
    def longest_inference_time(self):
        if self.in_flight_request_start_times == []:
            return 0

        oldest_start_time = min([start_time for _, start_time in self.in_flight_request_start_times])

        return time.time() - oldest_start_time

    def update(self, event) -> "PotassiumStatus":
        event_type = event[0]
        event_data = event[1:]
        if event_type not in event_handlers:
            raise InvalidStatusEvent(f"Invalid status event: {event_type}")
        return event_handlers[event_type](self.clone(), *event_data)


    def clone(self):
        return PotassiumStatus(
            self.num_started_inference_requests,
            self.num_completed_inference_requests,
            self.num_bad_requests,
            self.num_workers,
            self.num_workers_started,
            self.idle_start_timestamp,
            self.in_flight_request_start_times
        )

def handle_start_inference(status: PotassiumStatus, request_id: RequestID):
    status.in_flight_request_start_times.append((request_id, time.time()))
    return status

def handle_end_inference(status: PotassiumStatus, request_id: RequestID):
    status.num_completed_inference_requests += 1
    status.in_flight_request_start_times = [t for t in status.in_flight_request_start_times if t[0] != request_id]

    if status.gpu_available:
        status.idle_start_timestamp = time.time()

    return status

def handle_inference_request_received(status: PotassiumStatus):
    status.num_started_inference_requests += 1
    return status

def handle_worker_started(status: PotassiumStatus):
    status.num_workers_started += 1
    return status

def handle_bad_request_received(status: PotassiumStatus):
    status.num_bad_requests += 1
    return status

event_handlers = {
    StatusEvent.INFERENCE_REQUEST_RECEIVED: handle_inference_request_received,
    StatusEvent.INFERENCE_START: handle_start_inference,
    StatusEvent.INFERENCE_END: handle_end_inference,
    StatusEvent.WORKER_STARTED: handle_worker_started,
    StatusEvent.BAD_REQUEST_RECEIVED: handle_bad_request_received
}





from enum import Enum
import time
import os
from types import GeneratorType
from typing import Callable
from dataclasses import dataclass
from flask import Flask, request, make_response, abort, Response as FlaskResponse
import uuid
from werkzeug.serving import make_server
from threading import Thread, Lock
from queue import Queue as ThreadQueue
import functools
from termcolor import colored
from multiprocessing import Pool as ProcessPool, Queue as ProcessQueue
from multiprocessing.pool import ThreadPool
from .status import PotassiumStatus, StatusEvent
from .worker import run_worker, init_worker
from .exceptions import RouteAlreadyInUseException, InvalidEndpointTypeException
from .types import Request, RequestHeaders, Response
import logging

class HandlerType(Enum):
    HANDLER = "HANDLER"
    BACKGROUND = "BACKGROUND"

@dataclass
class Endpoint():
    type: HandlerType
    func: Callable

class ResponseMailbox():
    def __init__(self, response_queue):
        self._response_queue = response_queue
        self._mailbox = {}
        self._lock = Lock()

        t = Thread(target=self._response_handler, daemon=True)
        t.start()

    def _response_handler(self):
        try:
            while True:
                request_id, payload = self._response_queue.get()
                with self._lock:
                    if request_id not in self._mailbox:
                        self._mailbox[request_id] = ThreadQueue()
                    self._mailbox[request_id].put(payload)
        except EOFError:
            # queue closed, this happens when the server is shutting down
            pass

    def get_response(self, request_id):
        with self._lock:
            if request_id not in self._mailbox:
                self._mailbox[request_id] = ThreadQueue()
        result, stream_id = self._mailbox[request_id].get()

        if stream_id is not None:
            result.body = self._stream_body(stream_id)

        with self._lock:
            del self._mailbox[request_id]

        return result

    def _stream_body(self, stream_id):
        with self._lock:
            if stream_id not in self._mailbox:
                self._mailbox[stream_id] = ThreadQueue()
        queue = self._mailbox[stream_id]

        try:
            while True:
                result = queue.get()
                if isinstance(result, Exception):
                    with self._lock:
                        del self._mailbox[stream_id]
                    raise result
                elif result == None:
                    break
                else:
                    yield result
        except GeneratorExit:
            while True:
                # flush the queue
                result = queue.get()
                if result == None:
                    break
                elif isinstance(result, Exception):
                    with self._lock:
                        del self._mailbox[stream_id]
                    raise result
        with self._lock:
            del self._mailbox[stream_id]


class Potassium():
    "Potassium is a simple, stateful, GPU-enabled, and autoscaleable web framework for deploying machine learning models."

    def __init__(self, name, experimental_num_workers=1):
        self.name = name

        # default init function, if the user doesn't specify one
        self._init_func = lambda _: {}
        # dictionary to store unlimited Endpoints, by unique route
        self._endpoints = {}  
        self._context = {}
        self._flask_app = self._create_flask_app()
        self._event_queue = ProcessQueue()
        self._response_queue = ProcessQueue()
        self._response_mailbox = ResponseMailbox(self._response_queue)

        self._num_workers = experimental_num_workers

        self._worker_pool = None

        self.event_handler_thread = Thread(target=self._event_handler, daemon=True)
        self.event_handler_thread.start()

        self._status = PotassiumStatus.initial(self._num_workers)

    def _event_handler(self):
        try:
            while True:
                event = self._event_queue.get()
                self._status = self._status.update(event)
        except EOFError:
            # this happens when the process is shutting down
            pass


    def init(self, func):
        """init runs once on server start, and is used to initialize the app's context.
        You can use this to load models onto the GPU, set up connections, etc.
        The context is then passed into all handlers and background tasks.
        Notes:
        - this function must return a dictionary
        - the context is not persisted to disk, and will be reloaded on server restart
        - the context is not shared between multiple replicas of the app
        """

        self._init_func = func
        return func
    
    @staticmethod
    def _standardize_route(route):
        # handle empty or None case
        if len(route) == 0 or route == None:
            route = "/"

        # prepend with "/" if not already, as router expects
        if route[0] != "/":
            route = "/" + route
        
        return route

    def _base_decorator(self, route: str, handler_type: HandlerType):
        route = self._standardize_route(route)
        if route in self._endpoints:
            raise RouteAlreadyInUseException()

        def actual_decorator(func):
            @functools.wraps(func)
            def wrapper(context, request):
                # send in app's stateful context if GPU, and the request
                out = func(context, request)

                if handler_type == HandlerType.HANDLER:
                    if type(out) != Response:
                        raise Exception("Potassium Response object not returned")
                    if type(out.body) != bytes and type(out.body) != GeneratorType:
                        raise Exception(
                            "Potassium Response object body must be bytes", type(out.body))

                return out

            
            self._endpoints[route] = Endpoint(type=handler_type, func=wrapper)
            return wrapper
        return actual_decorator

    # handler is a blocking http POST handler
    def handler(self, route: str = "/"):
        "handler is a blocking http POST handler"
        return self._base_decorator(route, HandlerType.HANDLER)

    # background is a non-blocking http POST handler
    def background(self, route: str = "/"):    
        "background is a non-blocking http POST handler"
        return self._base_decorator(route, HandlerType.BACKGROUND)

    def test_client(self):
        "test_client returns a Flask test client for the app"
        self._init_server()
        return self._flask_app.test_client()

    def _create_flask_app(self):
        flask_app = Flask(__name__)

        # ingest into single endpoint and spread out to multiple downstream funcs
        @flask_app.route('/', defaults={'path': ''}, methods=["POST"])
        @flask_app.route('/<path:path>', methods=["POST"])
        def handle(path):
            route = "/" + path
            if route not in self._endpoints:
                self._event_queue.put((StatusEvent.BAD_REQUEST_RECEIVED,))
                abort(404)

            endpoint = self._endpoints[route]
            request_id = request.headers.get("X-Banana-Request-Id", None)
            if request_id is None:
                request_id = str(uuid.uuid4())
            try:
                req = Request(
                    headers=RequestHeaders(dict(request.headers.items())),
                    json=request.get_json(),
                    id=request_id
                )
            except:
                res = make_response()
                res.status_code = 400
                self._event_queue.put((StatusEvent.BAD_REQUEST_RECEIVED,))
                return res

            self._event_queue.put((StatusEvent.INFERENCE_REQUEST_RECEIVED,))

            assert self._worker_pool is not None, "Worker pool not initialized"
            # use an internal id for critical path to prevent user from accidentally
            # breaking things by sending multiple requests with the same id
            internal_id = str(uuid.uuid4())
            if endpoint.type == HandlerType.HANDLER:
                self._worker_pool.apply_async(run_worker, args=(endpoint.func, req, internal_id, True))
                resp = self._response_mailbox.get_response(internal_id)

                flask_response = FlaskResponse(
                    resp.body,
                    status=resp.status,
                    headers=resp.headers
                )
            elif endpoint.type == HandlerType.BACKGROUND:
                self._worker_pool.apply_async(run_worker, args=(endpoint.func, req, internal_id))

                flask_response = make_response({'started': True})
            else:
                raise InvalidEndpointTypeException()

            return flask_response

        @flask_app.route('/_k/warmup', methods=["POST"])
        def warm():
            request_id = str(uuid.uuid4())

            # a bit of a hack but we need to send a start and end event to the event queue
            # in order to update the status the way the load balancer expects
            self._event_queue.put((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
            self._event_queue.put((StatusEvent.INFERENCE_END, request_id))
            res = make_response({
                "warm": True,
            })
            res.status_code = 200
            return res

        @flask_app.route('/_k/status', methods=["GET"])
        @flask_app.route('/__status__', methods=["GET"])
        def status():
            cur_status = self._status

            res = make_response({
                "gpu_available": cur_status.gpu_available,
                "sequence_number": cur_status.sequence_number,
                "idle_time": int(cur_status.idle_time*1000),
                "inference_time": int(cur_status.longest_inference_time*1000),
            })

            res.status_code = 200
            return res

        return flask_app
    
    def _init_server(self):
        # unless the user has already set up logging, set up logging to stdout using
        # a separate fd so that we don't get in the way of request logs
        log = logging.getLogger('werkzeug')
        if len(log.handlers) == 0:
            # duplicate stdout
            stdout_copy = os.dup(1)
            # redirect flask logs to stdout_copy
            log.addHandler(logging.StreamHandler(os.fdopen(stdout_copy, 'w')))

        self._idle_start_time = time.time()
        index_queue = ProcessQueue()
        for i in range(self._num_workers):
            index_queue.put(i)
        if self._num_workers == 1:
            Pool = ThreadPool
        else:
            Pool = ProcessPool
        self._worker_pool = Pool(
            self._num_workers,
            init_worker, 
            (
                index_queue,
                self._event_queue,
                self._response_queue, 
                self._init_func,
                self._num_workers
            )
        )

        while True:
            if self._status.num_workers_started == self._num_workers:
                break
        print(colored(f"Started {self._num_workers} workers", 'green'))

    # serve runs the http server
    def serve(self, host="0.0.0.0", port=8000):
        print(colored("------\nStarting Potassium Server 🍌", 'yellow'))
        self._init_server()
        server = make_server(host, port, self._flask_app, threaded=True)
        print(colored(f"Serving at http://{host}:{port}\n------", 'green'))

        server.serve_forever()




import pytest
from potassium.status import StatusEvent, PotassiumStatus, InvalidStatusEvent
import time

@pytest.mark.parametrize("worker_num", [
    1,
    2,
    4
])
def test_workers_starting(worker_num):
    status = PotassiumStatus.initial(worker_num)
    assert status.num_workers == worker_num
    assert status.num_workers_started == 0
    assert status.gpu_available == False
    status = status.update((StatusEvent.WORKER_STARTED,))

    if worker_num == 1:
        assert status.gpu_available == True
    else:
        assert status.gpu_available == False

    for _ in range(worker_num-1):
        status = status.update((StatusEvent.WORKER_STARTED,))
    assert status.num_workers_started == worker_num
    assert status.gpu_available == True

def test_bad_event():
    status = PotassiumStatus.initial(1)
    with pytest.raises(InvalidStatusEvent):
        status.update(("BAD_EVENT",))

def test_inference_requests_single_worker():
    status = PotassiumStatus.initial(1)
    status = status.update((StatusEvent.WORKER_STARTED,))

    status = status.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    assert status.num_started_inference_requests == 1
    assert status.num_completed_inference_requests == 0
    assert status.gpu_available == False

    status = status.update((StatusEvent.INFERENCE_START, 0))
    status = status.update((StatusEvent.INFERENCE_END, 0))

    assert status.num_started_inference_requests == 1
    assert status.num_completed_inference_requests == 1
    assert status.sequence_number == 1
    assert status.gpu_available == True

    status = status.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    assert status.num_started_inference_requests == 2
    status = status.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    assert status.num_started_inference_requests == 3

    status = status.update((StatusEvent.INFERENCE_START, 1))
    status = status.update((StatusEvent.INFERENCE_END, 1))

    assert status.num_started_inference_requests == 3
    assert status.num_completed_inference_requests == 2
    assert status.sequence_number == 3
    assert status.gpu_available == False

    status = status.update((StatusEvent.INFERENCE_START, 2))
    status = status.update((StatusEvent.INFERENCE_END, 2))

    assert status.num_started_inference_requests == 3
    assert status.num_completed_inference_requests == 3
    assert status.sequence_number == 3
    assert status.gpu_available == True

def test_inference_requests_multiple_workers():
    state = PotassiumStatus.initial(2)

    state = state.update((StatusEvent.WORKER_STARTED,))
    state = state.update((StatusEvent.WORKER_STARTED,))

    assert state.gpu_available == True
    assert state.sequence_number == 0

    state = state.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    assert state.num_started_inference_requests == 1
    assert state.num_completed_inference_requests == 0
    assert state.gpu_available == True

    state = state.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    assert state.num_started_inference_requests == 2
    assert state.num_completed_inference_requests == 0
    assert state.gpu_available == False

    state = state.update((StatusEvent.INFERENCE_START, 0))
    state = state.update((StatusEvent.INFERENCE_END, 0))

    assert state.num_started_inference_requests == 2
    assert state.num_completed_inference_requests == 1
    assert state.sequence_number == 2
    assert state.gpu_available == True

    state = state.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))
    state = state.update((StatusEvent.INFERENCE_REQUEST_RECEIVED,))

    assert state.num_started_inference_requests == 4
    assert state.num_completed_inference_requests == 1
    assert state.sequence_number == 4
    assert state.gpu_available == False

    state = state.update((StatusEvent.INFERENCE_START, 1))
    state = state.update((StatusEvent.INFERENCE_END, 1))

    assert state.num_started_inference_requests == 4
    assert state.num_completed_inference_requests == 2
    assert state.sequence_number == 4
    assert state.gpu_available == False

    state = state.update((StatusEvent.INFERENCE_START, 2))
    state = state.update((StatusEvent.INFERENCE_END, 2))

    assert state.num_started_inference_requests == 4
    assert state.num_completed_inference_requests == 3
    assert state.sequence_number == 4
    assert state.gpu_available == True

    state = state.update((StatusEvent.INFERENCE_START, 3))
    state = state.update((StatusEvent.INFERENCE_END, 3))

    assert state.num_started_inference_requests == 4
    assert state.num_completed_inference_requests == 4
    assert state.sequence_number == 4
    assert state.gpu_available == True

@pytest.mark.parametrize("status_result_tuple", [
    (PotassiumStatus(
        num_started_inference_requests=0,
        num_completed_inference_requests=0,
        num_bad_requests=0,
        num_workers=1,
        num_workers_started=0,
        idle_start_timestamp=0,
        in_flight_request_start_times=[]
    ), 0),
    (PotassiumStatus(
        num_started_inference_requests=0,
        num_completed_inference_requests=0,
        num_bad_requests=0,
        num_workers=1,
        num_workers_started=1,
        idle_start_timestamp=0,
        in_flight_request_start_times=[]
    ), time.time()),
    (PotassiumStatus(
        num_started_inference_requests=1,
        num_completed_inference_requests=0,
        num_bad_requests=0,
        num_workers=1,
        num_workers_started=1,
        idle_start_timestamp=0,
        in_flight_request_start_times=[]
    ), 0),
    (PotassiumStatus(
        num_started_inference_requests=2,
        num_completed_inference_requests=0,
        num_bad_requests=0,
        num_workers=4,
        num_workers_started=4,
        idle_start_timestamp=0,
        in_flight_request_start_times=[]
    ), 0),
    (PotassiumStatus(
        num_started_inference_requests=2,
        num_completed_inference_requests=2,
        num_bad_requests=0,
        num_workers=4,
        num_workers_started=4,
        idle_start_timestamp=0,
        in_flight_request_start_times=[]
    ), time.time()),
])
def test_idle_time(status_result_tuple):
    status, result = status_result_tuple
    delta = abs(status.idle_time - result)
    ALLOWED_DELTA = 1
    assert delta < ALLOWED_DELTA

def test_longest_inference_time():
    status = PotassiumStatus(
        num_started_inference_requests=6,
        num_completed_inference_requests=2,
        num_bad_requests=0,
        num_workers=4,
        num_workers_started=4,
        idle_start_timestamp=0,
        in_flight_request_start_times=[
            ("b", time.time() - 2),
            ("a", time.time() - 1),
            ("c", time.time() - 3),
            ("d", time.time()),
        ]
    )

    longest_inference_time = status.longest_inference_time
    EXPECTED_LONGEST_INFERENCE_TIME = 3
    delta = abs(longest_inference_time - EXPECTED_LONGEST_INFERENCE_TIME)

    ALLOWED_DELTA = 0.1
    assert delta < ALLOWED_DELTA







from multiprocessing import Queue
import os
import threading
from typing import Dict, Any, Generator
from dataclasses import dataclass
from flask import make_response, Response as FlaskResponse
from termcolor import colored
import traceback
import inspect

from .status import StatusEvent
from .types import Response

worker = None

class FDRedirect():
    def __init__(self, fd: int):
        self._fd = fd
        self._fd_copy = os.dup(fd)
        self._redirect_w = None

    def _run_redirect_loop(self, redirect_r, prefix):
        redirect_r = os.fdopen(redirect_r, "r")

        for line in redirect_r:
            os.write(self._fd_copy, (prefix + line).encode("utf-8"))
        redirect_r.close()

    def set_prefix(self, prefix):
        if self._redirect_w is not None:
            os.dup2(self._fd_copy, self._fd)
            os.close(self._redirect_w)

        fd = self._fd
        redirect_r, redirect_w = os.pipe()

        self._fd_copy = os.dup(fd)
        os.dup2(redirect_w, fd)
        self._redirect_w = redirect_w

        t = threading.Thread(target=self._run_redirect_loop, args=(redirect_r, prefix))
        t.daemon = True
        t.start()


@dataclass
class Worker():
    worker_num: int
    total_workers: int
    context: Dict[Any, Any]
    event_queue: Queue
    response_queue: Queue
    stderr_redirect: FDRedirect
    stdout_redirect: FDRedirect

def init_worker(index_queue, event_queue, response_queue, init_func, total_workers):
    global worker
    worker_num = index_queue.get()

    stdout_redirect = FDRedirect(1)
    stderr_redirect = FDRedirect(2)

    if total_workers > 1:
        stderr_redirect.set_prefix(f"[worker {worker_num}] ")
        stdout_redirect.set_prefix(f"[worker {worker_num}] ")

    # check if the init function takes in a worker number
    print(colored("Running init()", 'yellow'))
    try:
        if len(inspect.signature(init_func).parameters) == 0:
            context = init_func()
        else:
            context = init_func(worker_num)
    except Exception as e:
        tb_str = traceback.format_exc()
        print(colored(tb_str, "red"))
        raise e

    if not isinstance(context, dict):
        raise Exception("Potassium init() must return a dictionary")

    event_queue.put((StatusEvent.WORKER_STARTED,))

    worker = Worker(
        worker_num,
        total_workers,
        context,
        event_queue,
        response_queue,
        stdout_redirect,
        stderr_redirect
    )

def run_worker(func, request, internal_id, use_response=False):
    assert worker is not None, "worker is not initialized"
    
    if worker.total_workers > 1:
        prefix = f"[worker {worker.worker_num}, requestID {request.id}] "
    else:
        prefix = f"[requestID {request.id}] "


    worker.stderr_redirect.set_prefix(prefix)
    worker.stdout_redirect.set_prefix(prefix)

    resp = None
    worker.event_queue.put((StatusEvent.INFERENCE_START, internal_id))

    try:
        resp = func(worker.context, request)
    except:
        tb_str = traceback.format_exc()
        print(colored(tb_str, "red"))
        resp = Response(
            status=500,
            body=tb_str.encode("utf-8"),
            headers={
                "Content-Type": "text/plain"
            }
        )

    if use_response:
        generator = None
        stream_id = None
        if inspect.isgenerator(resp.body):
            stream_id = 'stream-' + internal_id
            generator = resp.body
            resp.body = None
        worker.response_queue.put((internal_id, (resp, stream_id)))

        # if the response is a generator, we need to iterate through it
        if stream_id:
            assert generator is not None
            for chunk in generator:
                worker.response_queue.put((stream_id, chunk))
            worker.response_queue.put((stream_id, None))


    if worker.total_workers == 1:
        worker.stderr_redirect.set_prefix("")
        worker.stdout_redirect.set_prefix("")

    worker.event_queue.put((StatusEvent.INFERENCE_END, internal_id))

